/* Programme en mai-juin 1996 par Jean-Francois Lozevis
   Emulateur Spectrum 48 ko classique
   pour Linux avec architecture 386 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>                                             
#include <malloc.h>

#include "compile.h"     
#include "hash.h"

int par=0,debogue=0;
char *name_file;
FILE *fout;
                
char *getParam(int);

void uef(void) {
	own_error("Unexpected end of file");
}
                
long getNum() {
	int c,sign;
	long n;
	          
	sign=0;     
	while (lastCar=='-' || isspace(lastCar)) {
		if (lastCar=='-')
			sign++;
		getCar();
	}
	if (lastCar<'0' || lastCar>'9')
		own_error("Digit expected");
	n=lastCar-'0';
	while(1) {
		switch (c=getCar()) {
			case '0':
			case '1':
			case '2':
			case '3':
			case '4':
			case '5':
			case '6':
			case '7':
			case '8':
			case '9':
				n=n*10+c-'0';
				break;
			case EOF:
				uef();
			default:
				lastCar=c;
				return sign&1?-n:n;
			}	
		}               
}
                
long getExpr(int flagGetcar);                
                
long calcOpt(long val) {
	while(1) {
		switch(lastCar) {
			case '\t':
			case '\n':
			case ' ':
				getCar();
				break;
			case '+':
				val+=getExpr(1);
				break;
			case '-':
				val-=getExpr(1);
				break;
			case '*':
				val*=getExpr(1);
				break;
			case '/':
				val/=getExpr(1);
				break;
			case '&':
				val&=getExpr(1);
				break;
			case '|':
				val|=getExpr(1);
				break;
			case '%':
				val%=getExpr(1);
				break;
			case '>':
				switch (getCar()) {
					case '=':
						val=val>=getExpr(1)?1:0;
						break;
					case '>':
						val>>=getExpr(1);
						break;
				 	default:
						val=val>getExpr(0)?1:0;
				}
				break;
			case '<':            
				switch (getCar()) {
					case '=':
						val=val<=getExpr(1)?1:0;
						break;
				    case '>':
						val=val!=getExpr(1)?1:0;
						break;
					case '<':
						val<<=getExpr(1);
						break;
				 	default:
						val=val<getExpr(0)?1:0;
				}
				break;
			case '=':
				val=val==getExpr(1)?1:0;
				break;
			case '!':
				if (getCar()!='=')
					own_error("= expected");
				val=val!=getExpr(1)?1:0;
				break;
			default:
				return val;
		}
	}
}

long EquExpr(void) {
	char *s1,*s2;
	long res;
	
	while (isspace(getCar()))
		;
	if (lastCar!='(')
		own_error("( expected");
	getCar();
	s1=getParam(',');
	getCar();
	s2=getParam(')');
	res=strcmp(s1,s2)?0:1;
	free(s1);
	free(s2);
	return res;
}

long getExpr(int flagGetcar) {
	long v=0;
	int c;
        
	while (1) {
		switch(c=flagGetcar?getCar():(flagGetcar=1,lastCar)) {
			case ' ':
			case '\t':
			case '\n':
				continue;
			case '=':
				return EquExpr();
			case '0':
			case '1':
			case '2':
			case '3':
			case '4':
			case '5':
			case '6':
			case '7':
			case '8':
			case '9':
			case '-':
				return getNum();
			case '\'':
				c=getCar();
				getCar();
				return c;
			case '(':
				par++;
				v=calcOpt(getExpr(1));
				par--;
				if (lastCar!=')')
					own_error(") expected");
				getCar();
				break;
			case ']':
				if (par>0)
					own_error(") expected, not [");
				break;
			case EOF:
				uef();
			default:
				own_error("Unknown expression");
		}
		return v;
	}
}			    

long calcExpr(void) {
	return calcOpt(getExpr(1));
}

void help(void) {
	printf("Takes two files in arguments:\n");
	printf("\tmacro text input file\n");
	printf("\tmacro text output file\n");
	printf("Options are:\n");
	printf("\t/h or /? : prints this help\n");
	printf("\t/l : lists all macros\n");
	printf("\t/d : adds debugging text\n");
	printf("\t/sSYMBOL : defines the symbol SYMBOL\n");
}

void defSymbol(char *str) {
	char *ptr;
	
	for (ptr=str;*ptr;ptr++)
		if (*ptr=='=') {
			*ptr=0;
			if (!ptr[1])
				break;
			hash_add(str,0,ptr+1);
			return;
		}
	hash_add(str,0," ");
}

int main(int argc,char **argv) {
	static int listing=0;

	int c,i,nParams=1;
	char *inputFile=NULL,*outputFile=NULL;
	            
//	fprintf(stderr,"Macro compiler for the script language\nWritten by Jean-Francois Lozevis\n");
	init_hash();
	for (i=1;i<argc;i++) {
		if (argv[i][0]=='-')
			switch (argv[i][1]) {
				case 'h':
				case '?':
					help();
					return 0;
				case 'l':
					listing=1;
					break;
				case 'd':
					debogue=1;
					break;
				case 's':
					defSymbol(argv[i]+2);
					break;
				default:
					fprintf(stderr,"Unknown option /%c\n",argv[i][1]);
					help();
					exit(2);
				}
		else if (inputFile==NULL)
			inputFile=argv[i];
		else if (outputFile==NULL)
			outputFile=argv[i];
		else {
			char *s=own_malloc(nParams>9?4:3);
			sprintf(s,"_%d",nParams++);
			hash_add(s,0,argv[i]);
		}
	}
	if (inputFile==NULL || outputFile==NULL) {
		fprintf(stderr,"Two files in argument are needed\n");
		exit(2);
	}              
	if ((!(fout=fopen(outputFile,"wt")))) {
		fprintf(stderr,"Can not create %s\n",outputFile);
		exit(2);
	}
	name_file=inputFile;
	init_macro();
	nbsymbols=0;
	open_file(inputFile);
	                                
	par=0;
	while ((c=getCar())!=EOF)
		fputc(c,fout);
	fprintf(stderr,"%d symbols defined\n",nbsymbols);
#ifdef _DOS
	fprintf(stderr,"%u bytes available, %u contiguous\n",_memavl(),_memmax());
#endif // _DOS
	fclose(fout);
	if (listing)
		printAllMacros();
	return 0;
}
