/* Programme en mai-juin 1996 par Jean-Francois Lozevis
   Emulateur Spectrum 48 ko classique
   pour Linux avec architecture 386 */
# include <stdio.h>
# include <string.h> 
# include <stdlib.h>                                                                       

# include "hash.h"

struct hash_base **hash_table;
unsigned line=1;
unsigned nbsymbols;     

void *own_malloc(size_t n) {
  void *ptr;
  
  ptr=(void *) malloc(n);
  if (!ptr)
    own_error("Out of memory");
  return ptr;
}
                   
void init_hash(void) {                                    
	int i;
	
	hash_table=own_malloc(SIZE_HASH*sizeof(struct hash_base *));
	for (i=0;i<SIZE_HASH;i++)
		hash_table[i]=NULL;
//	last_hash_table=hash_table+SIZE_HASH-1;
}

static unsigned signature(char *name) {
	unsigned n;
	
	n=name[0];
	while (*(++name))
		n=97*n+*name;
	return n;
}

static struct hash_base **hash_preseek(char *name) {
	struct hash_base **ptr;
	
	for (ptr=hash_table+(signature(name) & (SIZE_HASH-1));*ptr;ptr=&(*ptr)->next)
		if (!strcmp(name,(*ptr)->name))
			return ptr;
	return ptr;
}

void hash_add(char *name,int npar,char *str) {
	struct hash_base **ptr,*hptr;
	
	ptr=hash_preseek(name);
	if (*ptr) {
		fprintf(stderr,"symbol %s already defined",name);
		own_error(NULL);
	}           
	nbsymbols++;
	*ptr=hptr=own_malloc(sizeof(struct hash_base));
	hptr->name=name;
	hptr->npar=npar;
	hptr->str=str;
	hptr->next=NULL;
}

void hash_del(char *name) {
	struct hash_base **ptr,*hptr;
	
	ptr=hash_preseek(name);
	if (!(hptr=*ptr)) {
		fprintf(stderr,"Symbol %s not defined",name);
		own_error(NULL);
	}              
	if (hptr->npar<0)
		own_error("Key words are not erasable");
	nbsymbols--;
	*ptr=hptr->next; 
	free(hptr->name);
	free(hptr->str);
	free(hptr);
}

struct hash_base *hash_seek(char *name) {
	return *hash_preseek(name);
}

int hash_exist(char *name) {
	return *hash_preseek(name)?1:0;
}                      

void printAllMacros(void) {
	struct hash_base *ptr;
	int i,prof;
	
	for (i=0;i<SIZE_HASH;i++)
		if ((prof=0,ptr=hash_table[i])!=NULL)
			do {
				if (ptr->npar>=0)
					printf("Macro %s with %d parameters, prof: %d value:\n%s\n",
						ptr->name,ptr->npar,++prof,ptr->str);
				else
					printf("Predefined macro %s, number %d, prof: %d\n",
					ptr->name,-ptr->npar,++prof);
			}
			while ((ptr=ptr->next)!=NULL);
}
