/* Programme en mai-juin 1996 par Jean-Francois Lozevis
   Emulateur Spectrum 48 ko classique
   pour Linux avec architecture 386 */
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>

#include "spectrum.h"
#include "son.h"

extern char ZXiBase[1],ZXiCB[1],ZXiDDCB[1],ZXiFDCB[1],ZXiED[1],ZXiDD[1],
  ZXiFD[1];

#if DEBUG_IN
char debug_ins[65536];
#endif // DEBUG_IN


void InitDebogue(void) {
#if DEBUG_IN
  int i;

  for (i=0;i<65536;i++)
    debug_ins[i]=0;
#endif // DEBUG_IN
}

static char *SeekS(char *s,int i) {
  while (i>0) {
    i--;
    while (*s++)
      ;
  }
  return s;
}

static char *Unass(unsigned *adr) {
  unsigned PC=*adr;
  char *s;

  switch (mem[PC]) {
  case 0xCB:
    s=SeekS(ZXiCB,mem[PC+1]);
    break;
  case 0xED:
    s=SeekS(ZXiED,mem[PC+1]);
    break;
  case 0xFD:
    if (mem[PC+1]==0xCB)
      s=SeekS(ZXiFDCB,mem[PC+3]);
    else
      s=SeekS(ZXiFD,mem[PC+1]);
    break;
  case 0xDD:
    if (mem[PC+1]==0xCB)
      s=SeekS(ZXiDDCB,mem[PC+3]);
    else
      s=SeekS(ZXiDD,mem[PC+1]);
    break;
  default:
    s=SeekS(ZXiBase,mem[PC]);
    break;
  }
  *adr+=s[0];
  return s+1;
}

static unsigned DisplayZX(void) {
  unsigned SP=header_sna.regSPL+header_sna.regSPH*256;

  printf("A=%2x B=%2x C=%2x D=%2x E=%2x H=%2x L=%2x intMode=%1x\n",
	 header_sna.regA,
	 header_sna.regB,
	 header_sna.regC,
	 header_sna.regD,
	 header_sna.regE,
	 header_sna.regH,
	 header_sna.regL,
	 header_sna.intMode);
  printf("A'=%2x B'=%2x C'=%2x D'=%2x E'=%2x H'=%2x L'=%2x iff2=%x\n",
	 header_sna.regAp,
	 header_sna.regBp,
	 header_sna.regCp,
	 header_sna.regDp,
	 header_sna.regEp,
	 header_sna.regHp,
	 header_sna.regLp,
	 header_sna.iff2);
  printf("PC=%4x SP=%4x I=%2x F=%2x IX=%4x IY=%4x\n",
	 PC,SP,
	 header_sna.regI,
	 header_sna.regF,
	 header_sna.regLX+256*header_sna.regHX,
	 header_sna.regLY+256*header_sna.regHY);
  printf("[PC]=%2x,%2x,%2x,%2x (%s) [SP]=%2x,%2x,%2x,%2x\n",
	 mem[PC],
	 mem[PC+1],
	 mem[PC+2],
	 mem[PC+3],
	 Unass(&PC),
	 mem[SP],
	 mem[SP+1],
	 mem[SP+2],
	 mem[SP+3]);
  return PC;
}

static unsigned Hextoi(char *s) {
  unsigned r=0;
  int c;

  while (1) {
    c=toupper(*s);
    s++;
    if (c>='0' && c<='9')
      r=(r<<4)+(c-'0');
    else
      if (c>='A' && c<='F')
	r=(r<<4)+(c-'A'+10);
      else
	return r;
  }
}

static void ZXg(char *s) {
  unsigned poss[20];
  unsigned char vals[20];
  int i;

  for (i=0;i<20;i++) {
      while (isspace(*s))
	s++;
      if (*s==0)
	break;
      poss[i]=Hextoi(s);
      vals[i]=mem[poss[i]];
      mem[poss[i]]=0x76;
      while (*s && !isspace(*s))
	s++;
  }
  ExecZ80(1000000000);
  while (i>0) {
    i--;
    mem[poss[i]]=vals[i];
  }
}
	  

void debogue(void) {
  char s[80];
  int n;
  unsigned pt;

  pt=DisplayZX();
  while (1) {
    printf("ZX>");
    if (!fgets(s,79,stdin))
      exit(0);
    switch(s[0]) {
    case '\n':
      break;
    case 't':
	n=s[1]!='\n'?Hextoi(s+1):1;
	while (n>0) {
	  n--;
	  ExecZ80(0);
	}
	pt=DisplayZX();
	break;
    case 'q':
      exit(0);
    case 'h':
    case '?':
      printf("Aide:\n\tt[n] do 1 or n instructions\n\tq quit\n\t? ou h affiche cette aide\n\tu[n] desassemble 10 ou n instructions\n\tcn place le desassembleur a l'adresse n\n\tr demarre l'emulation\n\tgn [n*] emule jusqu'aux adresses\n");
#if DEBUG_IN
      printf("\tI debug_in\n");
#endif // DEBUG_IN
      break;
#if DEBUG_IN
    case 'I':
      for (n=0;n<65536;n++)
	if (debug_ins[n]) {
	  debug_ins[n]=0;
	  printf("[%4x]",n);
	}
      break;
#endif // DEBUG_IN
    case 'c':
      pt=Hextoi(s+1);
      break;
    case 'u':
      n=s[1]!='\n'?Hextoi(s+1):10;
      while (n>0) {
	printf("%4x: %2x %2x %2x %2x",
		pt,mem[pt],mem[pt+1],mem[pt+2],mem[pt+3]);
	printf("  %s\n",Unass(&pt));
	n--;
      }
      break;
    case 'g':
      ZXg(s+1);
      PC--;
      pt=DisplayZX();
      break;
    case 'r':
      return;
    default:
      printf("Commande inconnue: %c\n",s[0]>=' '?s[0]:' ');
      break;
    }
  }
}
