/* Programme en mai-juin 1996 par Jean-Francois Lozevis
   Emulateur Spectrum 48 ko classique
   pour Linux avec architecture 386 */
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xos.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdio.h>
#include <signal.h>
#include <sys/time.h>
#include <X11/keysym.h>
#include <stdlib.h>

#include "spectrum.h"

unsigned char attrsChanged[768];
extern Display *display;
extern int screen;
extern GC gc;
extern Window win;
XImage *ximageb,*ximagew,*ximage3;
unsigned char bytes_images[8];
unsigned char inv_bytes[256];
unsigned long words_images[8];
unsigned long l3_images[24];
unsigned long inv_words[256];
unsigned long inv_3[256];
unsigned long ZXcolors[16];
int Z80activite;
int ZXflash=0;
int ZXflashy;
int typeclavier=0;
unsigned PC=0;
unsigned char Z80Keys[8]={255,255,255,255,255,255,255,255};
int Gmode=1;
char *nameSna;

struct Header_Z80 {
  unsigned char regA,regF,regC,regB,regL,regH,regPCL,regPCH,regSPL,regSPH;
  unsigned char regI,regR,regZob,regE,regD,regCp,regBp,regEp,regDp,regLp,regHp;
  unsigned char regAp,regFp,regLY,regHY,regLX,regHX,regDIEI,regIff2,regIntMode;
};

struct Header_Z80_2 {
  unsigned short lAdd;
  unsigned char regPCL2,regPCH2,HardMode,zob1,zob2,zob3,zob4,zob5[16];
};

static void ErrLit(char *name) {
  fprintf(stderr,"Impossible de lire correctement %s\n",name);
  exit(1);
}

static void LitZ80(char *name) {
  FILE *f;
  struct Header_Z80 Z80;
  struct Header_Z80_2 Z80_2;
  int c,yaeuED=0;
  unsigned char *pt;

  if ((f=fopen(name,"rb"))==NULL) {
    fprintf(stderr,"Impossible d'ouvrir le fichier: %s\n",name);
    exit(1);
  }
  if (fread(&Z80,1,sizeof(Z80),f)!=sizeof(Z80))
    ErrLit(name);
  PC=Z80.regPCL+256*Z80.regPCH;
  if (PC==0) {
    if (fread(&Z80_2,1,sizeof(Z80_2),f)!=sizeof(Z80_2))
      ErrLit(name);
    if (Z80_2.HardMode!=0) {
      fprintf(stderr,"%s n'est pas un fichier Spectrum 48 Ko\n",name);
      exit(1);
    }
    PC=Z80_2.regPCL2+256*Z80_2.regPCH2;
  }
  header_sna.intMode=Z80.regIntMode&3;
  header_sna.regA=Z80.regA;
  header_sna.regB=Z80.regB;
  header_sna.regC=Z80.regC;
  header_sna.regD=Z80.regD;
  header_sna.regE=Z80.regE;
  header_sna.regH=Z80.regH;
  header_sna.regL=Z80.regL;
  header_sna.regAp=Z80.regAp;
  header_sna.regBp=Z80.regBp;
  header_sna.regCp=Z80.regCp;
  header_sna.regDp=Z80.regDp;
  header_sna.regEp=Z80.regEp;
  header_sna.regHp=Z80.regHp;
  header_sna.regLp=Z80.regLp;
  header_sna.regF=Z80.regF;
  header_sna.regFp=Z80.regFp;
  header_sna.regR=Z80.regR;
  header_sna.regI=Z80.regI;
  header_sna.regLX=Z80.regLX;
  header_sna.regHX=Z80.regHX;
  header_sna.regLY=Z80.regLY;
  header_sna.regHY=Z80.regHY;
  header_sna.regSPL=Z80.regSPL;
  header_sna.regSPH=Z80.regSPH;
  header_sna.iff2=Z80.regDIEI?4:0;
  pt=mem+16384;
  if (Z80.regZob&32)
    while ((c=fgetc(f))!=EOF)
      if (c==0xED) {
	c=fgetc(f);
	if (c==0xED) {
	  int i=fgetc(f);
	  if (!i) {
	    fprintf(stderr,"%d octets decompresses\n",pt-mem-16384);
	    break;
	  }
	  c=fgetc(f);
	  if (c==EOF)
	    ErrLit(name);
	  pt-=yaeuED;
	  while (i>0) {
	    *pt++=c;
	    i--;
	  }
	  yaeuED=0;
	}
	else {
	  *pt++=0xED;
	  *pt++=c;
	  yaeuED=1;
	}
      }
      else {
	*pt++=c;
	yaeuED=0;
      }
  else
    if (fread(pt,1,48*1024,f)!=48*1024)
      ErrLit(name);
  if (pt>mem+65537) {
    fprintf(stderr,"Mauvais fichier .z80: %s\n",name);
//    exit(1);
  }
  fclose(f);
}

static void sigAlrm(int z) {
  XClientMessageEvent event;

  signal(SIGALRM,sigAlrm);
  if (!Z80activite) {
    ZXflash++;
    Z80activite=1;
    if (Gmode>=0 && fullSpeed==0) {
      event.type=ClientMessage;
      event.send_event=1;
      event.display=display;
      event.window=win;
      event.message_type=0;
      event.format=32;
      XSendEvent(display,win,0,0,(XEvent *) &event);
      /* XFlush(display);*/
    }
  }
}

static void InitColors(void) {
  Colormap cm=Gmode>=0?DefaultColormap(display,screen):0;
  XColor xc;
  int i;
  /* RGB(0,0,0),RGB(0,0,255),RGB(255,0,0),RGB(255,0,255)
     RGB(0,255,0),RGB(0,255,255),RGB(255,255,0),RGB(255,255,255)} */
  static unsigned short reds[8]={0,0,0xffff,0xffff,0,0,0xffff,0xffff};
  static unsigned short greens[8]={0,0,0,0,0xffff,0xffff,0xffff,0xffff};
  static unsigned short blues[8]={0,0xffff,0,0xffff,0,0xffff,0,0xffff};

  for (i=0;i<16;i++) {
    xc.red=i<8?reds[i]:reds[i-8]>>1;
    xc.green=i<8?greens[i]:greens[i-8]>>1;
    xc.blue=i<8?blues[i]:blues[i-8]>>1;
    if (Gmode>=0) {
      if (!XAllocColor(display,cm,&xc)) {
	fprintf(stderr,"Impossible d'allouer la couleur %d\n",i);
	exit(1);
      }
      ZXcolors[i]=xc.pixel;
    }
  }
}

void LitSna(char *name) {
  int i,f;
  unsigned SP;

  if (name==NULL)
    header_sna.regSPH=254;
  else if (name[strlen(name)-1]=='0')
    LitZ80(name);
  else {
    if ((f=open(name,O_RDONLY))==-1) {
      fprintf(stderr,"Impossible d'ouvrir %s\n",name);
      exit(1);
    }
    if (read(f,&header_sna,27)!=27
	|| read(f,mem+16384,48*1024)!=48*1024) {
      fprintf(stderr,"Impossible de lire %s\n",name);
      exit(1);
    }
    close(f);
    SP=header_sna.regSPL+256*header_sna.regSPH;
    PC=mem[SP]+256*mem[SP+1];
    header_sna.regSPL=(unsigned char) (SP+2);
    header_sna.regSPH=(unsigned char) ((SP+2)>>8);
  }
  if (name)
    fprintf(stderr,"Snapshot %s lu avec succes\n",name);
  for (i=0;i<768;i++) {
    whichAttrsChanged[i+256*24]=attrsChanged+i;
    attrsChanged[i]=1;
  }
}

void EcritSna(char *name) {
  int f;
  unsigned SP=header_sna.regSPL+256*header_sna.regSPH;

  mem[SP-2]=(unsigned char) PC;
  mem[SP-1]=(unsigned char) (PC>>8);
  header_sna.regSPL=(unsigned char) (SP-2);
  header_sna.regSPH=(unsigned char) ((SP-2)>>8);
  if ((f=creat(name,-1))==-1) {
    fprintf(stderr,"Impossible d'ouvrir %s\n",name);
    exit(1);
  }
  if (write(f,&header_sna,27)!=27
      || write(f,mem+16384,48*1024)!=48*1024) {
    fprintf(stderr,"Impossible de lire %s\n",name);
    exit(1);
  }
  close(f);
  fprintf(stderr,"Snapshot %s ecrit avec succes\n",name);
  header_sna.regSPL=(unsigned char) (SP);
  header_sna.regSPH=(unsigned char) (SP>>8);
}

void setTimer(void) {
  struct itimerval tv={{0L,20000L*raffraichit},{0L,20000L*raffraichit}};
  setitimer(ITIMER_REAL,&tv,NULL);
}

void InitSpectrum(char *name) {
  int f,i,vb,vw,vl;
  int bbo=Gmode>=0?BitmapBitOrder(display):0;

  nameSna=name;
  if ((f=open("spectrum.rom",O_RDONLY))==-1) {
    fprintf(stderr,"Impossible d'ouvrir spectrum.rom\n");
    exit(1);
  }
  if (read(f,mem,16384)!=16384) {
    fprintf(stderr,"Impossible de lire spectrum.rom\n");
    exit(1);
  }
  close(f);
  mem[65536]=mem[0];
  LitSna(name);  
  for (i=0;i<256*24;i++)
    whichAttrsChanged[i]=attrsChanged+(i&255)+(i>>11<<8);
  for (i=0;i<256;i++) {
    for (vl=vb=vw=f=0;f<8;f++) {
      int n=bbo==LSBFirst?7-f:f;
      vb|=(i>>f)&1?1<<n:0;
      vw|=(i>>f)&1?3<<(n*2):0;
      vl|=(i>>f)&1?7<<(n*3):0;
    }
    inv_bytes[i]=vb;
    inv_words[i]=vw | (vw<<16);
    inv_3[i]=vl;
  }
  if (Gmode>=0) {
    ximageb=XCreateImage(display, DefaultVisual(display,screen), 1, XYBitmap,
			 0, bytes_images, 8, 8, 8, 1);
    ximagew=XCreateImage(display, DefaultVisual(display,screen), 1, XYBitmap,
			 0, (unsigned char *) words_images, 16, 16, 16, 2);
    ximage3=XCreateImage(display, DefaultVisual(display,screen), 1, XYBitmap,
			 0, (unsigned char *) l3_images, 24, 24, 32, 4);
  }
  InitColors();
  Z80activite=1;
  signal(SIGALRM,sigAlrm);
  setTimer();
}


static void DisplayX(void) {
  int x,y,i;
  unsigned char *bytes;
  unsigned col;

  for (y=0;y<24;y++)
    for (x=0;x<32;x++) {
      i=y*32+x;
      if (attrsChanged[i]) {
	attrsChanged[i]=0;
	col=mem[i+(16384+256*24)];
	if ((col&128) && (ZXflashy&1))
	  if (col&64) {
	    XSetBackground(display,gc,ZXcolors[col&7]);
	    XSetForeground(display,gc,ZXcolors[(col>>3)&7]);
	  }
	  else {
	    XSetBackground(display,gc,ZXcolors[(col&7)+8]);
	    XSetForeground(display,gc,ZXcolors[((col>>3)&7)+8]);
	  }
	else if (col&64) {
	  XSetForeground(display,gc,ZXcolors[col&7]);
	  XSetBackground(display,gc,ZXcolors[(col>>3)&7]);
	}
	else {
	  XSetForeground(display,gc,ZXcolors[(col&7)+8]);
	  XSetBackground(display,gc,ZXcolors[((col>>3)&7)+8]);
	}
	bytes=mem+16384+(i&255)+(i>>8<<11);
	switch(Gmode) {
	case 0:
	  bytes_images[0]=inv_bytes[bytes[0]];
	  bytes_images[1]=inv_bytes[bytes[1*256]];
	  bytes_images[2]=inv_bytes[bytes[2*256]];
	  bytes_images[3]=inv_bytes[bytes[3*256]];
	  bytes_images[4]=inv_bytes[bytes[4*256]];
	  bytes_images[5]=inv_bytes[bytes[5*256]];
	  bytes_images[6]=inv_bytes[bytes[6*256]];
	  bytes_images[7]=inv_bytes[bytes[7*256]];
	  XPutImage(display, win, gc, ximageb, 0, 0, x*8,
		    y*8, 8, 8);
	  break;
	case 1:
	  words_images[0]=inv_words[bytes[0]];
	  words_images[1]=inv_words[bytes[1*256]];
	  words_images[2]=inv_words[bytes[2*256]];
	  words_images[3]=inv_words[bytes[3*256]];
	  words_images[4]=inv_words[bytes[4*256]];
	  words_images[5]=inv_words[bytes[5*256]];
	  words_images[6]=inv_words[bytes[6*256]];
	  words_images[7]=inv_words[bytes[7*256]];
	  XPutImage(display, win, gc, ximagew, 0, 0, x*16,
		    y*16, 16, 16);
	  break;
	default:
	  l3_images[0*3]=l3_images[0*3+1]=l3_images[0*3+2]=inv_3[bytes[0]];
	  l3_images[1*3]=l3_images[1*3+1]=l3_images[1*3+2]=inv_3[bytes[1*256]];
	  l3_images[2*3]=l3_images[2*3+1]=l3_images[2*3+2]=inv_3[bytes[2*256]];
	  l3_images[3*3]=l3_images[3*3+1]=l3_images[3*3+2]=inv_3[bytes[3*256]];
	  l3_images[4*3]=l3_images[4*3+1]=l3_images[4*3+2]=inv_3[bytes[4*256]];
	  l3_images[5*3]=l3_images[5*3+1]=l3_images[5*3+2]=inv_3[bytes[5*256]];
	  l3_images[6*3]=l3_images[6*3+1]=l3_images[6*3+2]=inv_3[bytes[6*256]];
	  l3_images[7*3]=l3_images[7*3+1]=l3_images[7*3+2]=inv_3[bytes[7*256]];
	  XPutImage(display, win, gc, ximage3, 0, 0, x*24,
		    y*24, 24, 24);
	  break;
	}

      }
    }
}

void DisplaySpectrum(void) {
  int i;
  
  if (ZXflash>20) {
    ZXflash=0;
    for (i=0;i<768;i++)
      if (mem[i+(16384+24*256)]&128)
	attrsChanged[i]=1;
    ZXflashy++;
  }
  if (Gmode>=0)
    DisplayX();
}

void Z80HitKey(KeySym key,int bit) {
  static KeySym valKeys0[40]=
  {
    XK_Control_L,XK_z,XK_x,XK_c,XK_v,
    XK_a,XK_s,XK_d,XK_f,XK_g,
    XK_q,XK_w,XK_e,XK_r,XK_t,
    XK_1,XK_2,XK_3,XK_4,XK_5,
    XK_0,XK_9,XK_8,XK_7,XK_6,
    XK_p,XK_o,XK_i,XK_u,XK_y,
    XK_Return,XK_l,XK_k,XK_j,XK_h,
    XK_space,XK_Shift_L,XK_m,XK_n,XK_b
  };
  static KeySym valKeys1[40]=
  {
    XK_Control_L,XK_w,XK_x,XK_c,XK_v,
    XK_q,XK_s,XK_d,XK_f,XK_g,
    XK_a,XK_z,XK_e,XK_r,XK_t,
    XK_1,XK_2,XK_3,XK_4,XK_5,
    XK_0,XK_9,XK_8,XK_7,XK_6,
    XK_p,XK_o,XK_i,XK_u,XK_y,
    XK_Return,XK_l,XK_k,XK_j,XK_h,
    XK_space,XK_Shift_L,XK_m,XK_n,XK_b
  };
  KeySym *valKeys=typeclavier?valKeys1:valKeys0;
  int i,j;

  /* fprintf(stderr,"%d",bit);*/
  for (i=0;i<8;i++)
    for (j=1;j<32;j<<=1,valKeys++)
      if (*valKeys==key) {
	if (bit)
	  Z80Keys[i]|=j;
	else
	  Z80Keys[i]&=255-j;
	return;
      }
  if (bit)
    switch(key) {
    case XK_F9:
      fullSpeed^=1;
      break;
    case XK_F10:
      debogue();
      break;
    case XK_F2:
      EcritSna(nameSna);
      break;
    case XK_F3:
      LitSna(nameSna);
      break;
    case XK_F4:
      executeZ80++;
      break;
    }
}
